#include <mpi.h>
#include "hplssolver.h"
#include "global.h"
#include "globmat.h"
#include "hpglobal.h"
#include "seqfilesm.h"

#include <string.h>

/**
   function computes linear static problem on a macro-level
   material parameters are not defined in material models, they are
   obtained from a meso or micro level by a homogenization method
   
   JK, 4. 6. 2013
*/
void parallel_homogenization_linear_statics ()
{
  long i,j,combmin,combmax,ncomb,tid;
  double *lhs,*rhs,**condm,**condvlhs,**condvrhs;
//  double *fl,*fi;
  precond prec;  
//  char name[1100];
  char nameout[1100],namegid[1100],namedat[1100];
  
  strcpy (nameout,Outdm->outfn);
  strcpy (namegid,Outdm->outgrfn);
  strcpy (namedat,Outdm->outdiagfn);

  // *******************************************************************************************
  // *******************************************************************************************
  // **  all processors assemble, factorize and store matrices of each tile type
  // **  no communication is needed and it has to save a time
  // **  in future, a variant where matrices will be distributed via MPI could be implemented
  // *******************************************************************************************
  // *******************************************************************************************

  condm = new double* [Ntiletypes];
  condvlhs = new double* [Ntiletypes];
  condvrhs = new double* [Ntiletypes];

  // cycle over all types of tiles
  for (i=0;i<Ntiletypes;i++){
    //  array for the condensed matrices
    condm[i] = new double [Nbdof[i]*Nbdof[i]];
    memset (condm[i],0,(Nbdof[i]*Nbdof[i])*sizeof(double));
    
    //  array for the condensed vectors
    condvlhs[i] = new double [Nbdof[i]];
    memset (condvlhs[i],0,Nbdof[i]*sizeof(double));
    
    //  array for the condensed vectors
    condvrhs[i] = new double [Nbdof[i]];
    memset (condvrhs[i],0,Nbdof[i]*sizeof(double));
    
    //  condensation - elimination of all internal DOFs
    SSmat[i].condense (Gtm,condm[i],condvrhs[i],Slhs[i],Srhs[i],Nbdof[i],1,Out);
  }


  Ndofm=0;
  for (i=0;i<Ntiles;i++){
    for (j=0;j<Nbdoftiling[i];j++){
      if (Ndofm<Cnbn[i][j])
        Ndofm=Cnbn[i][j];
    }
  }
  
  gtopology *top;
  top = new gtopology;
  top->initiate (Cnbn,Nbdoftiling,Ntiles);


  if (Smat!=NULL)
    delete Smat;
  
  if (Mp!=NULL)
    delete Mp;
  Mp = new probdesc;
  

  Mp->ssle->tlinsol=ldl;
  Mp->tstorsm=skyline_matrix;
  Mp->tprob=linear_statics;
  
  Mp->straincomp = 1;
  Mp->strainpos  = 2;
  Mp->strainaver = 1;
  
  Mp->stresscomp = 1;
  Mp->stresspos  = 2;
  Mp->stressaver = 1;
  
  Mp->otherstate=0;
  
  Smat = new gmatrix;
  Smat->setval (Mp->ssle);
  Smat->initiate (top,Ndofm,Mp->tstorsm,1,Out);
  
  fprintf (stderr,"\n\n\n\n\n pocet DOFs %ld",Ndofm);
  fprintf (stderr,"\n ukladani %d",Smat->ts);
  fprintf (stderr,"\n solver Smat %d",Smat->tlinsol);
  fprintf (stderr,"\n solver Mp   %d",Mp->ssle->tlinsol);

  lhs = new double [Ndofm];
  rhs = new double [Ndofm];
  
  ncomb=Ncomb/Nproc;
  combmin=Myrank*ncomb;
  combmax=(Myrank+1)*ncomb;

  fprintf (stderr,"\n myrank %d   ncomb %ld  combmin %ld   combmax %ld",Myrank,ncomb,combmin,combmax);

  for (i=combmin;i<combmax;i++){
    Smat->sky->nullsky ();
    nullv (rhs,Ndofm);
    
    for (j=0;j<Ntiles;j++){
      //  tile id
      tid = Tiling[i][j];
      //  the last parameter is useless
      Smat->localized (condm[tid],Cnbn[j],Nbdoftiling[tid],0);
      //  localization of the right hand side
      locglob (rhs,condvrhs[tid],Cnbn[j],Nbdoftiling[tid]);
    }
    
    Smat->solve_system (top,prec,lhs,rhs,Out);
    //fprintf (stderr,"\n rhs %le     Smat %le  Ndofm %ld",rhs[Ndofm-1],Smat->sky->a[0],Ndofm);
    for (j=0;j<Ndofm;j++){
      fprintf (stdout,"glob lhs %ld   %le\n",j,lhs[j]);
    }
    
    for (long ii=0;ii<3;ii++){
      
      //  tile id
      tid = Tiling[i][ii];
      fprintf (stderr,"\n myrank %d   i %ld    tid %ld",Myrank,i,tid);
      globloc (lhs,condvlhs[tid],Cnbn[ii],Nbdoftiling[tid]);
      //  back substitution
      SSmat[tid].condense (top,condm[tid],condvlhs[tid],Slhs[tid],Srhs[tid],Nbdoftiling[tid],2,Out);
      
      for (j=0;j<Nidof[tid]+Nbdoftiling[tid];j++){
        Lsrs->lhs[j]=Slhs[tid][j];
        fprintf (stdout,"%ld lhs %ld   %le\n",ii,j,Slhs[tid][j]);
      }
    }
    
    /*
    //  names of output files
    strcpy (name,nameout);
    sprintf (&name[strlen (nameout)],"%d.out",i);
    strcpy (Outdm->outfn, name);
    if (Outdm->gf != grfmt_no)
      {
        strcpy (name,namegid);
        sprintf (&name[strlen (namegid)],"%d",Myrank+1);
        strcpy (Outdm->outgrfn, name);
      }
    if (Outdm->ndiag > 0)
      {
        strcpy (name,namedat);
        sprintf (&name[strlen (namedat)],"%d.dat",Myrank+1);
        strcpy (Outdm->outdiagfn, name);
      }  
    
    print_init(-1, "wt");
    
    Mm = Mmm;
    Mt = Mtt;
    //  computes required quantities
    compute_req_val (0);
    

    if (Outdm->nog.selnforce.st != sel_no)  // nodal forces are required
      {
        // rhs was rewritten by the solver call -> rebuild load vector
        mefel_right_hand_side (0,Lsrs->give_rhs(0),fl);
      }
    
    // in artificial time 0.0 the nodal forces contains only the load vector fl
    print_step(0, 0, 0.0, fl);
    
    // if the nodal forces are required for graphical output
    // reaction + load vector are printed in the artificial time 1.0
    if (Outdm->nog.selnforce.st != sel_no){
      //  indicator of strain computation
      //  it means, strains at integration points have not been computed
      Mp->strainstate=0;
      //  indicator of stress computation
      //  it means, stresses at integration points have not been computed
      Mp->stressstate=0;
      //  indicator of computation of other array
      //  it means, stresses at integration points have not been computed
      Mp->otherstate=0;
      // calculate nodal forces and reactions
      internal_forces(0, fi);
      
      // in artificial time 1.0 the nodal forces contains the load vector + reactions
      print_step(0, 1, 1.0, fi);
    }


    print_close();
    */
  }

}



