int main (int argc,char *argv[])
{
  XFILE *in1;
  XFILE *in2;
  siftop top1, top2;
  meshform meshfmt1, meshfmt2;

  fprintf (stdout,"\n\n *** MESH COMPARATOR ***\n");
  fprintf (stdout," --------------------------\n");
  if (argc < 3){
    fprintf (stderr,"\n Wrong number of command line parameters.");
    fprintf (stderr,"\n Use : meshcomp mesh1_file_name mesh2_file_name\n\n");
    delete Mp;  delete Mt;  delete Mm;  delete Mc;  delete Mb;
    return(1);
  }
  in1 = xfopen(argv[1],"r");
  if (in1 == NULL){
    fprintf (stderr,"\n Cannot open mesh file 1.");
    return(2);
  }
  in2 = xfopen(argv[2],"r");
  if (in2 == NULL){
    fprintf (stderr,"\n Cannot open mesh file 2.");
    return(2);
  }

  printf("Input format of mesh 1 (0=sifel/1=t3d: ")
  scanf("%d", &meshfmt1);
  printf("\nInput format of mesh 2 (0=sifel/1=t3d: ")
  scanf("%d", &meshfmt2);
  printf("\n");

  switch (meshfmt1)
  {
    case t3d:
      top1->import_t3d(in, 0);
      break;
    case sifel:
      ret = top1->read(in, 0, 0);
      break;
    default:
      print_err("unknown mesh format is required", __FILE__, __LINE__, __func__);
  }
  switch (meshfmt2)
  {
    case t3d:
      top2->import_t3d(in, 0);
      break;
    case sifel:
      top2->read(in, 0, 0);
      break;
    default:
      print_err("unknown mesh format is required", __FILE__, __LINE__, __func__);
  }
  
  printf("\n Comparing nodes ...");


  double dx, dy, dz, dr, err;
  double x1, y1, z1;
  long *n1, *n2, sn1, sn2;
  long i, j, k;

  err = 1e-8;
  n1 = new long[top1.nn];
  memset(n1, 0, sizeof(*n1)*top1.nn);
  n2 = new long[top2.nn];
  memset(n2, 0, sizeof(*n2)*top2.nn);

  for(i=0; i<top1.nn; i++)
  {
    x1 = top1.nodes[i].x; 
    y1 = top1.nodes[i].y; 
    z1 = top1.nodes[i].z; 
    for(j=0; j<top2.nn; j++)
    {
      dx = x1 - top2.nodes[j].x;
      dy = y1 - top2.nodes[j].y;
      dz = z1 - top2.nodes[j].z;
      dr = dx*dx + dy*dy + dz*dz;
      dr = sqrt(dr);
      if (dr < err)
      {
        n1[i] = j+1;
        n2[j] = i+1;
        break;
      }
    }
    printf("\r Comparing nodes ...%.1lf%%", double(i/top1.nn));
  }
  for(i=0, sn1=0; i<top1.nn; i++)
  {
    if (n1[i])  sn1++;
  }
  for(i=0, sn1=0; i<top2.nn; i++)
  {
    if (n2[i])  sn2++;
  }
  
  if ((sn1 == top1.nn) && (sn2 == top2.nn))
    printf(" Nodes are identical.\n");
  if ((sn1 == top1.nn) && (sn2 != top2.nn))
    printf(" Nodes of mesh1 are contained in mesh2.\n");
  if ((sn2 == top2.nn) && (sn1 != top1.nn))
    printf(" Nodes of mesh2 are contained in mesh1.\n");
  if ((sn2 != top2.nn) && (sn1 != top1.nn))
  {
    printf(" Meshes have different nodes.\n");
    return 0;
  }

  printf("\n Comparing elements ...");
  
  long nne1, fn, tfn;
  long *e1, *e2, se1, se2;

  e1 = new long[top1.ne];
  memset(e1, 0, sizeof(*e1)*top1.ne);
  e2 = new long[top2.nn];
  memset(e2, 0, sizeof(*e2)*top2.ne);
  for(i=0; i<top1.ne; i++)
  {
    nne1 = top1.elements[i].nne;
    for (j=0; j<top2.ne; j++)
    {
      if (nne1 == top2.elements[j].nne)
      {
        tfn = 0;
        for (k=0; k<nne1; k++)
        {
          fn = 0;
          for (l=0; l<nne1; l++)
          {
            if (top1.elements[i].nodes[k] == top2.elements[j].nodes[l])
            {
              fn = 1;
              tfn++;
              break;  
            }
          }
          if (fn == 0)
            break;
        }
        if (tfn == nne1)
        {  
          e1[i] = j+1;
          e2[j] = i+1;
          break;
        }
      }
    }
  }

  for(i=0, se1=0; i<top1.ne; i++)
  {
    if (e1[i])  se1++;
  }
  for(i=0, se1=0; i<top2.ne; i++)
  {
    if (e2[i])  se2++;
  }

  if ((se1 == top1.ne) && (se2 == top2.ne))
    printf(" Elements are identical.\n");
  if ((se1 == top1.ne) && (se2 != top2.ne))
    printf(" Mesh1 is subdomain of mesh2.\n");
  if ((se2 == top2.ne) && (se1 != top1.ne))
    printf(" Mesh2 is subdomain of mesh1.\n");
  if ((se2 != top2.ne) && (se1 != top1.ne))
    printf(" Meshes have different elements.\n");

  return 0;
}
